unit Unit1;

{$mode objfpc}{$H+}

interface

uses
	Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, Buttons, StdCtrls,

    BESEN, BESENValue, BESENObject, BESENErrors, BESENNativeObject, BESENObjectJSON,
    BESENConstants, BESENObjectGlobal, BESENNumberUtils, BESENObjectMath,

    BESENObjectFunctionPrototype,
    BESENObjectFunctionConstructor,
    BESENObjectFunctionArguments,
    BESENObjectFunction,

    BESENStringUtils,
    BESENStringTree,
    BESENStringList,
    BESENSelfBalancedTree,
    BESENScope,
    BESENRegExpCache,
    BESENRandomGenerator,
    BESENPointerSelfBalancedTree,
    BESENPointerList,
    BESENParser,
    BESENOpcodes,
    BESENObjectThrowTypeErrorFunction,
    BESENObjectStringPrototype,
    BESENObjectStringConstructor,
    BESENObjectString, BESENCharset;

    {$extendedsyntax on}
    {$writeableconst on}
    {$varstringchecks on}
    {$typeinfo on}


type

 { TScriptSystem }

 TScriptSystem = class
 public

 	// Execute code
 	procedure ImportScript(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var AResult:TBESENValue);
 	procedure Execute(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var AResult:TBESENValue);

 	// Prints text in command line
 	procedure Print(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
 	procedure alert(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
 	procedure NativePrompt(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
 	procedure NativeVersion(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
 	procedure NativeLog(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);

 end;

 { TActorInterface }

 TEntityHandle = packed record // mockup
	EntityIndex: integer;
end;

 TActorInterface = class(TBESENNativeObject)
 protected
 	procedure ConstructObject(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer); Override;
 	procedure FinalizeObject; virtual;
 public
     constructor Create(AInstance: TObject; APrototype: TBESENObject=nil; AHasPrototypeProperty: longbool=false); Overload; Override;
     procedure SetHandle(const in_handle: TEntityHandle);
 private
     eh: TEntityHandle;
 	function gethealth: single;
 	procedure sethealth(AValue: single);
     function ReadwantedLevel: integer;
     function getstate: integer;
 	procedure setstate(AValue: integer);
 	procedure SetWantedLevel(AValue: integer);
     //function grabfromhandle(): Tactor; inline;
     function GetDeadTime(): single;
 published
     procedure MakeActiveLocal(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
     procedure GiveWeapon(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
 	procedure StripWeapons(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
     procedure NetworkSync(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
     procedure SetActorAnimation(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
     procedure ExtinguishFires(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
     procedure IsActorLocal(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
     procedure WeaponData(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);

     procedure free(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
     procedure valid(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
     procedure move(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
 	property health: single read gethealth write sethealth;

     property WantedLevel: integer read ReadwantedLevel write SetWantedLevel;
     property State: integer read getstate write setstate;
     property DeadTime: single read GetDeadTime;

 end;

 { TGameInterface }

 TGameInterface = class(TBESENNativeObject)
 protected
 	procedure ConstructObject(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer); Override;
 	procedure FinalizeObject; virtual;
 public
     constructor Create(AInstance: TObject; APrototype: TBESENObject=nil; AHasPrototypeProperty: longbool=false); Overload; Override;
     destructor destroy; override;
 private
      FOnTick : TBESENObjectFunction;
      FOnCreatePlayer : TBESENObjectFunction;
 published
     procedure ResetCamera(const ThisArgument:TBESENValue;Arguments:PPBESENValues;CountArguments:integer;var ResultValue:TBESENValue);
     property OnTick : TBESENObjectFunction read FOnTick write FOnTick;
 	property OnCreatePlayer : TBESENObjectFunction read FOnCreatePlayer write FOnCreatePlayer;
 end;

	{ TForm1 }

 TForm1 = class(TForm)
		BitBtn1: TBitBtn;
		Memo1: TMemo;
		Memo2: TMemo;
		procedure BitBtn1Click(Sender: TObject);
  procedure FormCreate(Sender: TObject);
  procedure Memo2Change(Sender: TObject);
	private

	public

	end;

var
	Form1: TForm1;

    BesenInst: TBesen;
	ScriptSystem: TScriptSystem;

    objectlist: array of pointer;
    test_worked: boolean = false;

implementation

{$R *.lfm}

{ TGameInterface }

procedure TGameInterface.ConstructObject(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer);
begin
    setlength(objectlist, length(objectlist) + 1);

    objectlist[high(objectlist)]:= self;

	inherited ConstructObject(ThisArgument, Arguments, CountArguments);

end;

procedure TGameInterface.FinalizeObject;
begin

end;

constructor TGameInterface.Create(AInstance: TObject; APrototype: TBESENObject; AHasPrototypeProperty: longbool);
begin
	inherited Create(AInstance, APrototype, AHasPrototypeProperty);
end;

destructor TGameInterface.destroy;
begin
	inherited destroy;
end;

procedure TGameInterface.ResetCamera(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

{ TActorInterface }

procedure TActorInterface.ConstructObject(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer);
begin
	inherited ConstructObject(ThisArgument, Arguments, CountArguments);
end;

procedure TActorInterface.FinalizeObject;
begin

end;

constructor TActorInterface.Create(AInstance: TObject; APrototype: TBESENObject; AHasPrototypeProperty: longbool);
begin
	inherited Create(AInstance, APrototype, AHasPrototypeProperty);
end;

procedure TActorInterface.SetHandle(const in_handle: TEntityHandle);
begin

end;

function TActorInterface.gethealth: single;
begin

end;

procedure TActorInterface.sethealth(AValue: single);
begin

    form1.Memo1.lines.add('sethealth has worked, this is a proper object.');

    test_worked:= true;

end;

function TActorInterface.ReadwantedLevel: integer;
begin

end;

function TActorInterface.getstate: integer;
begin

end;

procedure TActorInterface.setstate(AValue: integer);
begin

end;

procedure TActorInterface.SetWantedLevel(AValue: integer);
begin

end;

function TActorInterface.GetDeadTime: single;
begin

end;

procedure TActorInterface.MakeActiveLocal(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

procedure TActorInterface.GiveWeapon(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

procedure TActorInterface.StripWeapons(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

procedure TActorInterface.NetworkSync(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

procedure TActorInterface.SetActorAnimation(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

procedure TActorInterface.ExtinguishFires(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

procedure TActorInterface.IsActorLocal(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

procedure TActorInterface.WeaponData(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

procedure TActorInterface.free(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

procedure TActorInterface.valid(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

procedure TActorInterface.move(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin

end;

{ TScriptSystem }

procedure TScriptSystem.ImportScript(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var AResult: TBESENValue);
begin
  AResult.ValueType:=bvtUNDEFINED;

end;

procedure TScriptSystem.Execute(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var AResult: TBESENValue);
begin
  AResult.ValueType:=bvtUNDEFINED;
end;

procedure TScriptSystem.Print(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
var
	i:integer;
	v:PBESENValue;
	fOutput:widestring;

	procedure writeit(s:widestring);
	begin
		fOutput:= fOutput + s;
	end;

begin

	fOutput:='';
	ResultValue.ValueType:=bvtUNDEFINED;

	for i:=0 to CountArguments-1 do begin
		v:=Arguments^[i];
		case v^.ValueType of
			bvtUNDEFINED:begin
				writeit('undefined');
			end;
			bvtNULL:begin
				writeit('null');
			end;
			bvtBOOLEAN:begin
				if v^.Bool then begin
					writeit('true');
				end else begin
					writeit('false');
				end;
			end;
			bvtNUMBER:begin
				writeit(BESENFloatToStr(v^.Num));
			end;
			bvtSTRING:begin
				writeit(v^.Str);
			end;
			bvtOBJECT:begin
				writeit(TBESEN(BesenInst).ToStr(v^) + ' ' + BesenInst.ToStr(BesenInst.JSONStringify(BESENObjectValue(v^.Obj))) );
			end;
			bvtREFERENCE:begin
				writeit('reference');
			end;
		end;
	end;

	Form1.Memo1.lines.add(BESENEncodeString(BESENUTF16ToUTF8(fOutput),UTF_8,BESENLocaleCharset));

end;

procedure TScriptSystem.alert(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin
  ResultValue.ValueType:=bvtUNDEFINED;
end;

procedure TScriptSystem.NativePrompt(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin
  ResultValue.ValueType:=bvtUNDEFINED;
end;

procedure TScriptSystem.NativeVersion(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin
  ResultValue.ValueType:=bvtUNDEFINED;
end;

procedure TScriptSystem.NativeLog(const ThisArgument: TBESENValue; Arguments: PPBESENValues; CountArguments: integer; var ResultValue: TBESENValue);
begin
  ResultValue.ValueType:=bvtUNDEFINED;
end;

{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);
begin

end;

procedure TForm1.Memo2Change(Sender: TObject);
begin

end;

procedure TForm1.BitBtn1Click(Sender: TObject);
var
	i: Integer;
	a: array[0..5] of PBESENValue;
	JsPlayer: TBESENValue;

	AResult: TBESENValue;
//	ta: Tactor;
	JsCreatedPlayer: TActorInterface;
begin

	test_worked:= false;

	BesenInst := TBesen.Create(COMPAT_JS); // JavaScript compability, affects a bit how objects are returned.

	BesenInst.RecursionLimit := 128;

	ScriptSystem := TScriptSystem.Create;

	// Register global functions

	TBESEN(BesenInst).ObjectGlobal.RegisterNativeFunction('importScript', @ScriptSystem.ImportScript, 1, []);
	TBESEN(BesenInst).ObjectGlobal.RegisterNativeFunction('execute', @ScriptSystem.Execute, 1, []);
	TBESEN(BesenInst).ObjectGlobal.RegisterNativeFunction('print', @ScriptSystem.Print, 1, []);
	TBESEN(BesenInst).ObjectGlobal.RegisterNativeFunction('alert', @ScriptSystem.alert, 1, []);
	TBESEN(BesenInst).ObjectGlobal.RegisterNativeFunction('prompt',@ScriptSystem.NativePrompt,1,[]);
	TBESEN(BesenInst).ObjectGlobal.RegisterNativeFunction('version',@ScriptSystem.NativeVersion,0,[]);

	// Register our classes
	BesenInst.RegisterNativeObject('Actor', TActorInterface);
	BesenInst.RegisterNativeObject('Game', TGameInterface);

	try
		BesenInst.Execute(Memo2.Lines.GetText());
	except
		on e: EBESENError do
			Form1.Memo1.lines.add(Format('%s ( %s | Line %d ): %s', [e.Name, 'buffer', TBESEN(BesenInst).LineNumber, e.Message]));
		on e: Exception do
			Form1.Memo1.lines.add(Format('%s ( %s | Line %d ): %s', ['Exception', 'buffer', TBESEN(BesenInst).LineNumber, e.Message]));
	end;




	// create an instance of class Tactor, then have the gameinterface class's function called + send actor object instance as an parameter

     for i:= 0 to high(objectlist) do begin

         if (Tobject(objectlist[i]).ClassType = TGameInterface) then with TGameInterface(objectlist[i]) do begin

 			if Assigned(OnCreatePlayer) then begin

 				TBESEN(BesenInst).GarbageCollector.Protect(OnCreatePlayer);

 			    try
 			       AResult.ValueType := bvtBOOLEAN;

                   // create a "javascript" object instance
                   JsCreatedPlayer:= TActorInterface.Create(BesenInst, TBESEN(BesenInst).ObjectPrototype, false);

 				   TBESEN(BesenInst).GarbageCollector.Add(JsCreatedPlayer);
                   TBESEN(BesenInst).GarbageCollector.Protect(TBESENObject(JsCreatedPlayer));

                   a[0]:= @JsPlayer;
                   JsPlayer := BESENObjectValue(TBESENObject(JsCreatedPlayer));

 			       OnCreatePlayer.Call(BESENObjectValue(OnCreatePlayer), @a, 2, AResult);

 			    except
 			       on e: EBESENError do
 			       begin
 			         form1.Memo1.lines.add('%s ( Line %d ): %s', [e.Name, TBESEN(BesenInst).LineNumber, e.Message]);
 			       end;

 			       on e: exception do
 			       begin
 			         form1.Memo1.lines.add(Format('%s ( Line %d ): %s', ['Exception', TBESEN(Instance).LineNumber, e.Message]));
 			       end;
 			    end;

 				//ResultValue := BESENBooleanValue(True);

 			end;

		 end;

	 end;

	if test_worked = false then
		form1.Memo1.lines.add('test has failed, TActorInterface.sethealth was not called.');

end;

end.

